import java.util.Map;

public interface BExpr extends Genable 
{ 	
	public boolean isConst(); 
	public boolean isVar(); 
	public boolean isConstTrue(Map<String, Integer> consts);
	public boolean isConstFalse(Map<String, Integer> consts);

	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals);
}

class Not implements BExpr
{
	BExpr b;
	public Not(BExpr b) 
	{if (b==null) throw new ParserException("Argument to Not null");
	this.b=b;}
	
	public String genCode(Context con)
	{
		return "Not." + b.genCode(con);
	}

	public String toString()
	{
		return "!(" + b + ")";
	}
	
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		return "!(" + b.printWithVals(ivals,bvals) + ")";
	}
	
	
	public boolean isConstTrue(Map<String, Integer> consts) {return b.isConstFalse(consts);}
	public boolean isConstFalse(Map<String, Integer> consts) {return b.isConstTrue(consts);}
	public boolean isConst() {return b.isConst();}
	public boolean isVar() {return false;}
	
}

class And implements BExpr
{
	BExpr b1,b2;
	public And(BExpr b1, BExpr b2)
	{
		if (b1==null) throw new ParserException("LH argument to And null");
		if (b2==null) throw new ParserException("RH argument to And null");
		this.b1=b1; 
		this.b2=b2;
	}
	
	public String genCode(Context con)
	{
		return "And." + b1.genCode(con) + "." + b2.genCode(con);
	}

	public String toString()
	{
		return b1 + " && " + b2;
	}
	
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		return b1.printWithVals(ivals,bvals) + "&&" + b2.printWithVals(ivals,bvals);
	}
	
	public boolean isConstTrue(Map<String, Integer> consts) {return b1.isConstTrue(consts) && b2.isConstTrue(consts);}
	public boolean isConstFalse(Map<String, Integer> consts) 
	{
		return b1.isConst() && b2.isConst() && !isConstTrue(consts);
	}
	
	public boolean isConst() {return b1.isConst() && b2.isConst();}
	public boolean isVar() {return false;}
	
}

class Or implements BExpr
{
	BExpr b1,b2;
	public Or(BExpr b1, BExpr b2)
	{
		if (b1==null) throw new ParserException("LH argument to Or null");
		if (b2==null) throw new ParserException("RH argument to Or null");
		this.b1=b1; 
		this.b2=b2;
	}
	
	public String genCode(Context con)
	{
		return "Or." + b1.genCode(con) + "." + b2.genCode(con);
	}

	public String toString()
	{
		return b1 + " || " + b2;
	}
	
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		return b1.printWithVals(ivals,bvals) + "||" + b2.printWithVals(ivals,bvals);
	}
	

	public boolean isConstTrue(Map<String, Integer> consts) 
	{
		return b1.isConst() && b2.isConst() && (b1.isConstTrue(consts) || b2.isConstTrue(consts));
	}
	public boolean isConstFalse(Map<String, Integer> consts) {return b1.isConstFalse(consts) && b2.isConstFalse(consts);}
	
		
	public boolean isConst() {return b1.isConst() && b2.isConst();}
	public boolean isVar() {return false;}
	
	public int numIEs() {return 0;}
	public int numBEs()
	{
		if (isConst())
			return 0;
		else return 1;
	}
}
	
class Eq implements BExpr
{
	IExpr i1,i2;
	public Eq(IExpr i1, IExpr i2)
	{
		if (i1==null) throw new ParserException("LH argument to Eq null");
		if (i2==null) throw new ParserException("RH argument to Eq null");
		this.i1=i1; 
		this.i2=i2;
	}
	
	public String genCode(Context con)
	{
		return "Eq." + i1.genCode(con) + "." + i2.genCode(con);
	}

	public String toString()
	{
		return i1 + " = " + i2;
	}
	
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		return i1.printWithVals(ivals) + " = " + i2.printWithVals(ivals);
	}
	
	
	public boolean isConstTrue(Map<String, Integer> consts) 
	{
		return (isConst() && (i1.constVal(consts) == i2.constVal(consts)));
	}
	public boolean isConstFalse(Map<String, Integer> consts)
	{
		return (isConst() && !isConstTrue(consts) );
	}
	
	public boolean isConst() { return i1.isConst() && i2.isConst();}
	public boolean isVar() {return false;}
}

class Gt implements BExpr
{
	IExpr i1,i2;
	public Gt(IExpr i1, IExpr i2)
	{
		if (i1==null) throw new ParserException("LH argument to Gt null");
		if (i2==null) throw new ParserException("RH argument to Gt null");
		this.i1=i1; 
		this.i2=i2;
	}
	
	public String genCode(Context con)
	{
		return "Gt." + i1.genCode(con) + "." + i2.genCode(con);
	}

	public String toString()
	{
		return i1 + " > " + i2;
	}
	
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		return i1.printWithVals(ivals) + " > " + i2.printWithVals(ivals);
	}
	
	
	public boolean isConstTrue(Map<String, Integer> consts) 
	{
		return (isConst() && (i1.constVal(consts) > i2.constVal(consts)));
	}
	public boolean isConstFalse(Map<String, Integer> consts)
	{
		return (isConst() && !isConstTrue(consts) );
	}
	
	public boolean isConst() { return i1.isConst() && i2.isConst();}
	public boolean isVar() {return false;}
}

class Ge implements BExpr
{
	IExpr i1,i2;
	public Ge(IExpr i1, IExpr i2)
	{
		if (i1==null) throw new ParserException("LH argument to Ge null");
		if (i2==null) throw new ParserException("RH argument to Ge null");
		this.i1=i1; 
		this.i2=i2;

	}
	
	public String genCode(Context con)
	{
		return "Ge." + i1.genCode(con) + "." + i2.genCode(con);
	}
	
	public String toString()
	{
		return i1 + " >= " + i2;
	}
	
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		return i1.printWithVals(ivals) + " >= " + i2.printWithVals(ivals);
	}
	
	public boolean isConstTrue(Map<String, Integer> consts) 
	{
		return (isConst() && (i1.constVal(consts) >= i2.constVal(consts)));
	}
	public boolean isConstFalse(Map<String, Integer> consts)
	{
		return (isConst() && !isConstTrue(consts) );
	}
	
	
	public boolean isConst() { return i1.isConst() && i2.isConst();}
	public boolean isVar() {return false;}
}

class Lt implements BExpr
{
	IExpr i1,i2;
	public Lt(IExpr i1, IExpr i2)
	{
		if (i1==null) throw new ParserException("LH argument to Lt null");
		if (i2==null) throw new ParserException("RH argument to Lt null");
		this.i1=i1; 
		this.i2=i2;
	}
	
	public String genCode(Context con)
	{
		return "Gt." + i2.genCode(con) + "." + i1.genCode(con);
	}

	public String toString()
	{
		return i1 + " < " + i2;
	}
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		return i1.printWithVals(ivals) + " < " + i2.printWithVals(ivals);
	}
	
	
	public boolean isConstTrue(Map<String, Integer> consts) 
	{
		return (isConst() && (i1.constVal(consts) < i2.constVal(consts)));
	}
	public boolean isConstFalse(Map<String, Integer> consts)
	{
		return (isConst() && !isConstTrue(consts) );
	}
		
		
	public boolean isConst() { return i1.isConst() && i2.isConst();}
	public boolean isVar() {return false;}
}

class Le implements BExpr
{
	IExpr i1,i2;
	public Le(IExpr i1, IExpr i2)
	{
		if (i1==null) throw new ParserException("LH argument to Le null");
		if (i2==null) throw new ParserException("RH argument to Le null");
		this.i1=i1; 
		this.i2=i2;
	}
	
	public String genCode(Context con)
	{
		return "Ge." + i2.genCode(con) + "." + i1.genCode(con);
	}

	public String toString()
	{
		return i1 + " <= " + i2;
	}
	
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		return i1.printWithVals(ivals) + " <= " + i2.printWithVals(ivals);
	}
	
	public boolean isConstTrue(Map<String, Integer> consts) 
	{
		return (isConst() && (i1.constVal(consts) <= i2.constVal(consts)));
	}
	public boolean isConstFalse(Map<String, Integer> consts)
	{
		return (isConst() && !isConstTrue(consts) );
	}
	
	
	public boolean isConst() { return i1.isConst() && i2.isConst();}
	public boolean isVar() {return false;}
}

	
class True implements BExpr
{
	public final int ies = 0;
	public final int bes = 0;
	public String genCode(Context con) {return "True";}	

	public String toString()
	{ return "true";}
	
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		return "true";
	}
	
	public boolean isConstTrue(Map<String, Integer> consts) {return true;}
	public boolean isConstFalse(Map<String, Integer> consts){return false;}
	public boolean isConst() {return true;}
	public boolean isVar() {return false;}
}
class False implements BExpr
{
	public final int ies = 0;
	public final int bes = 0;
	public String genCode(Context con) {return "False";}

	
	public String toString()
	{ return "false";}
	
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		return "false";
	}
	
	public boolean isConstTrue(Map<String, Integer> consts) {return false;}
	public boolean isConstFalse(Map<String, Integer> consts){return true;}
	public boolean isConst() {return true;}
	public boolean isVar() {return false;}
}

class BVar implements BExpr
{
	String name;
	public BVar(String name)
	{
		if (name==null) throw new ParserException("Argument to Bvar null");
		this.name = name;
	}
	
	public String genCode(Context con) 
	{ 
		return "BVar." + con.getBEncoding(name); //con should throw an exception if name is not in con
	}

	public String toString() {return name;}
	
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		return bvals.containsKey(name)? name + "(" + bvals.get(name) + ")" : name;
	}
	
	public boolean isConstTrue(Map<String, Integer> consts) {return false;}
	public boolean isConstFalse(Map<String, Integer> consts){return false;}
	public boolean isConst() {return false;}
	public boolean isVar() {return true;}
}

class BArray implements BExpr
{
	String name;
	IConst i;
	public BArray(String name, IConst i)
	{
		if (name==null) throw new ParserException("name argument to BArray null"); 
		if (i==null) throw new ParserException("index argument to BArray null"); 
		
		this.name = name; this.i = i;
	}

	public String genCode(Context con) 
	{ 
		return "BVar." + con.getBAEncoding(name) + "." + i.genCode(con);
	}

	
	public String toString()
	{ return name + "[" + i + "]";}
	
	public String printWithVals(Map<String, Integer> ivals, Map<String,Boolean> bvals)
	{
		String me = name + "[" + i.constVal(ivals) + "]";
	
		if (bvals.containsKey(me)) return name + "[" + i.printWithVals(ivals) + "](" + bvals.get(me) + ")";
		else return name + "[" + i.printWithVals(ivals) + "]";
	}
	
	public boolean isConstTrue(Map<String, Integer> consts) {return false;}
	public boolean isConstFalse(Map<String, Integer> consts){return false;}
	public boolean isConst() {return false;}
	public boolean isVar() {return true;}
}