import java.io.*;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Map;

public class Interpreter
{
	static final boolean DEBUG = true;
	
	Parser parser;	
	BufferedReader reader = null;//reads the trace output from fdr

	public Interpreter(String source, String CSPFile, String fdrReadOutputFile, String fdrWriteOutputFile) throws FileNotFoundException, InterruptedException, IOException
	{
		this(source);
		if (fdrReadOutputFile != null)
		{
			loadOutput(fdrReadOutputFile);
		}
		else
		{
			runFDR(CSPFile,fdrWriteOutputFile);
		}
	}


	public Interpreter(String source) throws FileNotFoundException
	{
		trace("Creating parser for " + source);
		parser = new Parser(new FileReader(source));
		parser.yyparse();
	}
	
	public void loadOutput(String fdrReadOutputFile) throws FileNotFoundException
	{
		trace("Reading FDR output from " + fdrReadOutputFile);
		reader = new BufferedReader( new FileReader(fdrReadOutputFile));
	}
	
	public void runFDR(String CSPFile, String fdrWriteOutputFile) throws FileNotFoundException, IOException, InterruptedException
	{
		trace("Creating " + CSPFile);
		parser.output(new PrintStream(CSPFile));
		
		java.lang.Runtime runtime = Runtime.getRuntime();
		
		trace("Running FDR");
		java.lang.Process p = runtime.exec(new String[] {"fdr2", "batch", "-trace", CSPFile});
		reader = new BufferedReader (new InputStreamReader(p.getInputStream()));
		BufferedReader errorReader = new BufferedReader (new InputStreamReader(p.getErrorStream()));
		
		int result = p.waitFor();
		trace("FDR returns" + result);
		if (result != 0)
		{
			String s = "Error Running FDR:";
			while (errorReader.ready())
			{
				s = s + errorReader.readLine() + "\n";
			}
			throw new InterpreterException(s);
		}
		
		trace("FDR finshed");
		
		if (fdrWriteOutputFile != null)
		{
			trace("Outputting to " + fdrWriteOutputFile);
			PrintStream outputfile = new PrintStream(fdrWriteOutputFile);
			while (reader.ready())
			{
				outputfile.println(reader.readLine());
			}
			outputfile.close();
			
			reader = new BufferedReader( new FileReader(fdrWriteOutputFile));
		}
	}
	/*public Interpreter(String source, String fdrOutputFile) throws FileNotFoundException
	{
		parser = new Parser(new FileReader(source));
		parser.yyparse();
		
		reader = new BufferedReader( new FileReader(fdrOutputFile));
	}
	
	public Interpreter(String source, boolean writeFDROutput)throws FileNotFoundException, InterruptedException, IOException
	{this(source,"./output.csp",writeFDROutput,"./fdroutput");}
	
	public Interpreter(String source, String CSPFile, boolean writeFDROutput, String FDROutputFile) throws FileNotFoundException, InterruptedException, IOException
	{	
		parser = new Parser(new FileReader(source));
		parser.yyparse();
		parser.output(new PrintStream( CSPFile ));
		
		java.lang.Runtime runtime = Runtime.getRuntime();
		
		/*if (writeFDROutput)
		{
		
			
		
			
			String[] cmds = new String[] {"fdr2", "batch", "-trace", CSPFile, ">", FDROutputFile};
			if (DEBUG)
			{
				String cmd = "";
				for (int i = 0; i < cmds.length; i++)
					cmd = cmd + cmds[i] + " ";
				trace("executing: " + cmd);
			}
			java.lang.Process p = runtime.exec(cmds);
			p.waitFor();
			reader = new BufferedReader( new FileReader(FDROutputFile));
			
			
		} * /
		else
		{
			java.lang.Process p = runtime.exec(new String[] {"fdr2", "batch", "-trace", CSPFile});
			reader = new BufferedReader (new InputStreamReader(p.getInputStream()));
			p.waitFor();
			
			//the above commented out bit didnt write the output file correctly
			//so do it this way:
			if (writeFDROutput)
			{
				PrintStream outputfile = new PrintStream(FDROutputFile);
				while (reader.ready())
				{
					outputfile.println(reader.readLine());
				}
				outputfile.close();
				
				reader = new BufferedReader( new FileReader(FDROutputFile));
			}
						
		}
	}*/


	public void interpret(InterpreterOutput output) throws IOException
	{
		if (reader == null)
			throw new InterpreterException("Trying to interpret file before results are known");
		
		Iterator<Assertion> asserts = parser.asserts.iterator();
		Assertion next =  (asserts.hasNext())? asserts.next() : null;	

		while(reader.ready())
		{
			String s = reader.readLine();
			if (isStartOf(next, s)) 
			{ 
				interpretAssert(next, output);
				next =  (asserts.hasNext())? asserts.next() : null;
			}
			else output.uninterpreted(s);
		}
		
	}

	boolean isStartOf(Assertion assertion, String line)
	{
		if ( !line.startsWith("Checking ") || assertion == null )
			return false;
		String s = line.substring(9).replaceAll(" ", "");//remove "Checking " and any spaces
		String t = assertion.genCode(parser.context).substring(7).replaceAll(" ", ""); //remove "assert " and any spaces
		trace("Checking if " + s + " matches " + t);
		return (s.equals(t));
	}
	
	void interpretAssert(Assertion assertion, InterpreterOutput output)  throws IOException
	{
		if (!reader.ready()) throw new InterpreterException("interpreter.interpretAssert: end of stream reached");
		trace("Interpreting " + assertion);
		String s = reader.readLine();
		if (s.equals("true"))
		{
			output.startAssertion(assertion, true);
			return;
		}
		else if (s.equals("xfalse"))
		{
			output.startAssertion(assertion, false);
			
			Context con = parser.context;
			while (true)
			{
			
				if (!reader.ready()) throw new InterpreterException("interpreter.interpretAssert: end of stream reached");

				s = reader.readLine();
				String[] ss = s.split("\\.");
				trace("interpreting " + s);
					
				if (s.startsWith("END TRACE"))
					{output.endTrace();break;}
				else if (s.startsWith("_tau")){trace("Skipping _tau");}
				else if (ss[0].equals("bvevalT"))
				{
					if (ss[1].equals("B"))
						output.bVarEval(con.getBName(Integer.parseInt(ss[2])),Boolean.parseBoolean(ss[3]),Integer.parseInt(ss[4]));
					else if (ss[1].equals("BA"))
						output.bVarEval(con.getBAName(Integer.parseInt(ss[2]))+ "[" + ss[3] + "]",Boolean.parseBoolean(ss[4]),Integer.parseInt(ss[5]));
					else
						throw new InterpreterException("Line in unexpected format");
				}
				else if (ss[0].equals("bvwriteT"))
				{
					if (ss[1].equals("B"))
						output.bVarWrite(con.getBName(Integer.parseInt(ss[2])),Boolean.parseBoolean(ss[3]),Integer.parseInt(ss[4]));
					else if (ss[1].equals("BA"))
						output.bVarWrite(con.getBAName(Integer.parseInt(ss[2]))+ "[" + ss[3] + "]",Boolean.parseBoolean(ss[4]),Integer.parseInt(ss[5]));
					else
						throw new InterpreterException("Line in unexpected format");
				}
				else if (ss[0].equals("ivevalT"))
				{
					if (ss[1].equals("I"))
						output.iVarEval(con.getIName(Integer.parseInt(ss[2])),Integer.parseInt(ss[3]),Integer.parseInt(ss[4]));
					else if (ss[1].equals("IA"))
						output.iVarEval(con.getIAName(Integer.parseInt(ss[2]))+ "[" + ss[3] + "]",Integer.parseInt(ss[4]),Integer.parseInt(ss[5]));
					else
						throw new InterpreterException("Line in unexpected format");
				}
				else if (ss[0].equals("ivwriteT"))
				{
					if (ss[1].equals("I"))
						output.iVarWrite(con.getIName(Integer.parseInt(ss[2])),Integer.parseInt(ss[3]),Integer.parseInt(ss[4]));
					else if (ss[1].equals("IA"))
						output.iVarWrite(con.getIAName(Integer.parseInt(ss[2]))+ "[" + ss[3] + "]",Integer.parseInt(ss[4]),Integer.parseInt(ss[5]));
					else
						throw new InterpreterException("Line in unexpected format");
				}
				else if (ss[0].equals("breq"))
				{
					output.bExprStart(Integer.parseInt(ss[2]));
				}
				else if (ss[0].equals("ireq"))
				{
					output.iExprStart(Integer.parseInt(ss[2]));
				}
				else if (ss[0].equals("beval"))
				{
					output.bExprEnd(Integer.parseInt(ss[2]),Boolean.parseBoolean(ss[3]));
				}
				else if (ss[0].equals("ieval"))
				{
					output.iExprEnd(Integer.parseInt(ss[2]),Integer.parseInt(ss[3]));
				}
				else if (ss[0].equals("start_atom"))
				{
					output.startAtom(Integer.parseInt(ss[1]));
				}
				else if (ss[0].equals("end_atom"))
				{
					output.endAtom(Integer.parseInt(ss[1]));
				}
				else if (con.isSig(ss[0].substring(0, ss[0].length() -1)) || ss[0].equals("outofrangeT") || ss[0].equals("assertionfailedT") || ss[0].equals("noopT"))//we have to hide the t with which signals are tagged
				{
					output.signal(ss[0].substring(0, ss[0].length() -1), Integer.parseInt(ss[1]));
				}
				else if (con.isISig(ss[0].substring(0, ss[0].length() -1)))
				{
					output.iSignal(ss[0].substring(0, ss[0].length() -1),Integer.parseInt(ss[1]),Integer.parseInt(ss[2]));
				}
				/*else if (ss[0].equals("outofrange"))
				{
					output.outofrange();
				}
				else if (ss[0].equals("assertionfailedt"))
				{
					output.assertionfailed();
				}*/
				else trace("ignoring " + s);
			}
			//catch (Throwable t)
			//{ throw new InterpreterException("Error " + t.getMessage() + " encountered while interpreting line " + s);}			
		}
	}
		
	/*public String run(String source){return run(source,"output.csp");}
	public String run(String source, String parserOutput)
	{
		Parser parser;
		// parse a file
		parser = new Parser(new FileReader(source));
		PrintStream stream = new PrintStream(parserOutput);
		parser.output(stream);
		stream.close();
	
		java.lang.Runtime runtime = Runtime.getRuntime();
		java.lang.Process p = runtime.exec(new String[] {"fdr2", "batch", "-trace", parserOutput});
		BufferedReader r = new BufferedReader (new InputStreamReader(p.getInputStream()));
		p.waitFor();
		while (r.ready())
		{
			String s = r.readLine();
			if (s.startsWith("BEGIN TRACE"))
			{
				Interpreter interpreter = new Interpreter();
				interpreter.interpret(r,parser.procs,parser.context, System.out);
			}
			else trace(s);
		}
		

	}*/

	static void trace(String s)
	{
		if (DEBUG) System.out.println("Interpreter: " + s);
	}

}