import java.util.TreeMap;
import java.util.Map;
import java.util.Vector;
import java.util.Iterator;

class TableDataInterpreter implements InterpreterOutput
{
	boolean useHTML = false;
	
	protected Map<String, Integer> consts;
	
	protected Assertion current;
	protected Vector<Entry> currentTrace;
	protected Map<Integer, Map<String, Integer>> ivals;//maps processes to a map from variable to their values
	protected Map<Integer, Map<String, Boolean>> bvals;//maps processes to a map from variable to their values
	
	
	protected Vector<AssertionOutput> results;
	
	public class AssertionOutput //represents the interpreted output of a single assertion check
	{
		public final String assertion;
		public final boolean result;
		public final Vector<String> procs;
		public final Vector<Entry> counterExample;
	
		public AssertionOutput(Assertion assertion, boolean result, Vector<Entry> counterExample)
		{
			if (assertion==null)
				throw new InterpreterException("TableDataInterpreter.AssertionOutput given null assertion");
			if (!result && counterExample.size() == 0)
				throw new InterpreterException("TableDataInterpreter.AssertionOutput given empty counterexample for false assertion");
			
				
			this.assertion = assertion.toString();
			this.result = result;
			this.counterExample = counterExample;
			
			procs = new Vector<String>();
			for (int i = 0; i < assertion.numProcs(); i++)
				procs.add(assertion.process(i));
		}
		
	}	
	
	public class Entry //represents a row in a table of a counterexample
	{
		public final String text;
		public final int col;
		public boolean isVarRead = false;
		public boolean isExprStart = false;
		public boolean isAtomic = false;
		public Entry(String text, int col){this.text=text; this.col = col;}
		
	}
	
	public void startAssertion(Assertion assertion, boolean result)
	{
	
		if (result)
		{
			results.add(new AssertionOutput(assertion,true,new Vector<Entry>()));
		}
		else
		{
			current = assertion;
			currentTrace = new Vector<Entry>();
		}
	}
	
	public void endTrace()
	{
		results.add(new AssertionOutput(current,false,currentTrace));
		current = null;
		currentTrace = null;
	}
	
	public void bVarEval(String name, boolean value, int proc)
	{ 
		if (bvals.containsKey(Integer.valueOf(proc)))
		//this process is in the middle of calculating the value of an expression
		{
			bvals.get(Integer.valueOf(proc)).put(name, Boolean.valueOf(value));
			Entry e = new Entry(html(bold(name) + " is " + value), proc);
			e.isVarRead = true;
			currentTrace.add(e);
		}
		else //this variable just appears on its own, so do not add to bvals or flag as a variable read
			currentTrace.add(new Entry(html(bold(name) + " is " + value), proc));
	}
	
	public void iVarEval(String name, int value, int proc)
	{ 
		if (ivals.containsKey(Integer.valueOf(proc)))
		//this process is in the middle of calculating the value of an expression
		{
			ivals.get(Integer.valueOf(proc)).put(name, Integer.valueOf(value));
			Entry e = new Entry(html(bold(name) + " is " + value), proc);
			e.isVarRead = true;
			currentTrace.add(e);
		}
		else //this variable just appears on its own, so do not add to ivals or flag as a variable read
			currentTrace.add(new Entry(html(bold(name) + " is " + value), proc));
	}
	
	public void bVarWrite(String name, boolean value, int proc)
	{ currentTrace.add(new Entry(html(bold(name) + " assigned " + value), proc)); }
	
	public void iVarWrite(String name, int value, int proc)	
	{ currentTrace.add(new Entry(html(bold(name) + " assigned " + value), proc)); }
	
	
	public void iExprStart(int exp)
	{
	
		Entry e = new Entry(html("Evaluation of " + bold(current.getIExpr(exp, consts)) + " starts"), current.indexOfIExprOwner(exp,consts));
		e.isExprStart = true;
		currentTrace.add(e);
		
		//create a new map to keep track of the values of varaibles as they return
		//since IExprs never contain BExprs we dont need to deal with them
		ivals.put(Integer.valueOf(current.indexOfIExprOwner(exp,consts)), new TreeMap<String,Integer>());
	}
	public void bExprStart(int exp)
	{
		Entry e = new Entry(html("Evaluation of " + bold(current.getBExpr(exp, consts, emptyBMap)) + " starts"), current.indexOfBExprOwner(exp,consts));
		e.isExprStart = true;
		currentTrace.add(e);
		
		
		//create new maps to keep track of the values of varaibles as they return
		//need 2 since BExprs can contains integer and boolean variables
		ivals.put(Integer.valueOf(current.indexOfBExprOwner(exp,consts)), new TreeMap<String,Integer>());
		bvals.put(Integer.valueOf(current.indexOfBExprOwner(exp,consts)), new TreeMap<String,Boolean>());
		
	}
	
	public void iExprEnd(int exp, int result)
	{
		Map<String, Integer> theIVals = ivals.get(Integer.valueOf(current.indexOfIExprOwner(exp,consts)));
		theIVals.putAll(consts);
	
		currentTrace.add(new Entry(html(bold(current.getIExpr(exp, theIVals)) + " is " + result), current.indexOfIExprOwner(exp,consts)));
	
		//now we have finished evaluating this expressin we remove its entry
		ivals.remove(Integer.valueOf(current.indexOfIExprOwner(exp,consts)));	
	}
	
	public void bExprEnd(int exp, boolean result)
	{
		Map<String, Integer> theIVals = ivals.get(Integer.valueOf(current.indexOfBExprOwner(exp,consts)));
		theIVals.putAll(consts);
		Map<String, Boolean> theBVals = bvals.get(Integer.valueOf(current.indexOfBExprOwner(exp,consts)));

	
		currentTrace.add(new Entry(html(bold(current.getBExpr(exp, theIVals, theBVals)) + " is " + result),current.indexOfBExprOwner(exp,consts)));
		
		//now we have finished evaluating this expressin we remove its entries
		ivals.remove(Integer.valueOf(current.indexOfBExprOwner(exp,consts)));
		bvals.remove(Integer.valueOf(current.indexOfBExprOwner(exp,consts)));		
	}
	
	
	public void startAtom(int proc)
	{ 
		Entry e = new Entry("Atomic section entered", proc); 
		e.isAtomic = true;
		currentTrace.add(e);
	}
	
	public void endAtom(int proc)
	{ 
		Entry e = new Entry("Atomic section left", proc); 
		e.isAtomic = true;
		currentTrace.add(e);
	}
	
	
	public void signal(String signal, int proc)
	{ currentTrace.add(new Entry(signal,proc)); }	
	
	public void iSignal(String signal, int value, int proc)
	{ currentTrace.add(new Entry(signal + "." + value, proc)); }

	public void uninterpreted(String line)
	{/*stream.println("Uninterpreted: " + line);*/}

	public TableDataInterpreter(Map<String, Integer> consts)
	{	
		if (consts == null) throw new InterpreterException("TableDataInterpreter given null consts");
		this.consts = consts;
		ivals = new TreeMap<Integer, Map<String, Integer>>();
		bvals = new TreeMap<Integer, Map<String, Boolean>>();
		
		results = new Vector<AssertionOutput>();
	}
	
	public Vector<AssertionOutput> getResults() {return results;}

	String bold(String s)
	{
		if (useHTML) 
			return "<b>" + s + "</b>";
		else
			return s;
	}
	String html(String s)
	{
		if (useHTML) 
			return "<html>" + s + "</html>";
		else
			return s;
	}	
}